/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef BIT_H
#define BIT_H

#include <glib.h>
#include "square.h"
#include "init.h"

#define shiftDown(b)      ((b)>>8)
#define shift2Down(b)     ((b)>>16)
#define shiftUp(b)        ((b)<<8)
#define shift2Up(b)       ((b)<<16)
#define shiftRight(b)     (((b)<<1)&NOT_FILE_A)
#define shift2Right(b)    (((b)<<2)&NOT_FILE_AB)
#define shiftLeft(b)      (((b)>>1)&NOT_FILE_H)
#define shift2Left(b)     (((b)>>2)&NOT_FILE_GH)
#define shiftUpLeft(b)    (((b)<<7)&NOT_FILE_H)
#define shiftUpRight(b)   (((b)<<9)&NOT_FILE_A)
#define shiftDownLeft(b)  (((b)>>9)&NOT_FILE_H)
#define shiftDownRight(b) (((b)>>7)&NOT_FILE_A)

#define getNumberBitsGTEQ(bits, rank)	numberOfBits64(bits >> (rank<<3))
#define getNumberBitsLEEQ(bits, rank)	numberOfBits64(bits << ((7-rank)<<3))
#define getNumberBitsGT(bits, rank)		numberOfBits64(bits >> ((rank+1)<<3))
#define getNumberBitsLE(bits, rank)		numberOfBits64(bits << ((8-rank)<<3))
#define getNumberBitsEQ(bits, rank)		numberOfBits64(bits & BRANK[rank])

void chessBitPrint(guint64 x);

static inline gint getFirstBit64(guint64 x){
	register gint index;
	#ifdef INTEL_64
	__asm__(
	"bsf %1,%%rcx\n\t"
	:"=c" (index)
	:"r" (x)
	);
	#endif
	#ifdef INTEL_32
	__asm__(
	"bsf %1,%%ecx\n\t"
	"jnz 1f\n\t"
	"bsf %2,%%ecx\n\t"
	"add $32,%%ecx\n"
	"1:\n"
	:"=c" (index)
	:"r" (x), "r"(x>>32)
	);
	#endif
	#ifndef INTEL_64
	#ifndef INTEL_32
	register guint64 b = x & -x;
	index =  0;
	if (!(b & 0xffffffff)) { b >>= 32; index |= 32; }
	if (!(b & 0xffff)) { b >>= 16; index |= 16; }
	if (!(b & 0xff)) { b >>= 8; index |= 8; }
	if (!(b & 0xf)) { b >>= 4; index |= 4; }
	if (!(b & 0x3)) { b >>= 2; index |= 2; }
	if (!(b & 0x1)) { index |= 1; }
	#endif
	#endif
	return index;
}

static inline gint getLastBit64(guint64 x){
	gint index;
	#ifdef INTEL_64
	__asm__(
	"bsr %1,%%rcx\n\t"
	:"=c" (index)
	:"r" (x)
	);
	#endif
	#ifdef INTEL_32
	__asm__(
	"bsr %2,%%ecx\n\t"
	"jnz 1f\n\t"
	"bsr %1,%%ecx\n\t"
	"jmp 2f\n\t"
	"1:\n"
	"add $32,%%ecx\n"
	"2:\n"
	:"=c" (index)
	:"r" (x), "r"(x>>32)
	);
	#endif
	#ifndef INTEL_32
	#ifndef INTEL_64
	index = 63;
	if (!(x & 0xffffffff00000000ull)) { x <<= 32;	index -= 32; }
	if (!(x & 0xffff000000000000ull)) { x <<= 16;	index -= 16; }
	if (!(x & 0xff00000000000000ull)) { x <<= 8; index -= 8; }
	if (!(x & 0xf000000000000000ull)) { x <<= 4; index -= 4; }
	if (!(x & 0xc000000000000000ull)) { x <<= 2; index -= 2; }
	if (!(x & 0x8000000000000000ull)) { x <<= 1; index -= 1; }
	#endif
	#endif
	return index;
}

static inline gint getFirstBitAndClear64(guint64* x){
	guint index = getFirstBit64(*x);
	*x ^= 1ull<<index;
	return index;
}

static inline guint numberOfBits64(guint64 n){
    return bitsIn16bits[n & 0xffffu] + bitsIn16bits[(n >> 16) & 0xffffu] + 
	   bitsIn16bits[(n >> 32) & 0xffffu] + bitsIn16bits[(n >> 48) & 0xffffu];
}

static inline gboolean bitsGTEQ(guint64 bits, gint rank){
	if(bits >> (rank<<3)) return TRUE;
	return FALSE;
}

static inline gboolean bitsLEEQ(guint64 bits, gint rank){
	if(bits << ((7-rank)<<3)) return TRUE;
	return FALSE;
}

static inline gboolean bitsGT(guint64 bits, gint rank){
	if(bits >> ((rank+1)<<3)) return TRUE;
	return FALSE;
}

static inline gboolean bitsLE(guint64 bits, gint rank){
	if(bits << ((8-rank)<<3)) return TRUE;
	return FALSE;
}

static inline gboolean bitsEQ(guint64 bits, gint rank){
	if(bits & BRANK[rank]) return TRUE;
	return FALSE;
}

static inline guint64 getBitFile(guint64 x, gint file){
	if(file < 0 || file > 7) return 0;
	return x & BFILE[file];
}

static inline guint64 getBitRank(guint64 x, gint rank){
	if(rank < 0 || rank > 7) return 0;
	return x & BRANK[rank];
}

#endif
