/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <string.h>
#include <glib/gprintf.h>
#include <glib.h>
#include <gmodule.h>

#include "egbb.h"
#include "bit.h"

#ifndef WIN

#ifdef m64
#define EGBBDLL "egbbso64.so"
#endif

#ifndef m64
#define EGBBDLL "/egbbso.so"
#endif

gchar bitbasesPath[256] = "./egbb/";
#endif

#ifdef WIN
#define EGBBDLL "\\egbbdll.dll"
gchar bitbasesPath[256] = ".\\egbb\\";
#endif

guint32 egbbCacheSize = 16; 

#define _NOTFOUND 99999
#define WIN_SCORE 4500

enum {_EMPTY,_WKING,_WQUEEN,_WROOK,_WBISHOP,_WKNIGHT,_WPAWN,
		_BKING,_BQUEEN,_BROOK,_BBISHOP,_BKNIGHT,_BPAWN};
enum {LOAD_NONE,LOAD_4MEN,SMART_LOAD,LOAD_5MEN};
static gint egbb_load_type = SMART_LOAD;
typedef void (*PLOAD_EGBB) (gchar* path,gint cache_size,gint load_options);
typedef gint (*PPROBE_EGBB) (gint player, gint w_king, gint b_king,
							gint piece1, gint square1,
							gint piece2, gint square2,
							gint piece3, gint square3);
guint bbhits;
PPROBE_EGBB probe_egbb;
GModule *module = NULL;

void egbbLoadLibrary(){
	const gchar dll[] = EGBBDLL;
	gchar dllPath[256];
	guint newEgbbCacheSize = egbbCacheSize * 1048576;

	strcpy(dllPath,bitbasesPath);
	strcat(dllPath,dll);

	PLOAD_EGBB load_egbb;

	if(module != NULL) g_module_close(module);
	module = g_module_open (dllPath, G_MODULE_BIND_LAZY);
	if(!module){
		g_printf("info string %s (%s)\n",g_module_error(),dllPath);
		module = NULL;
		probe_egbb = NULL;
		return;
	}
	if(!g_module_symbol(module,"load_egbb_5men",(gpointer)&load_egbb)){
		g_error("%s\n",g_module_error());
	}
	if(!load_egbb){
		g_error("Symbol load_egbb_5men is NULL\n");
	}
	if(!g_module_symbol(module,"probe_egbb_5men",(gpointer)&probe_egbb)){
		g_error("%s\n",g_module_error());
	}
	if(!probe_egbb){
		g_error("Symbol load_egbb_5men is NULL\n");
	}
	load_egbb((gchar *)bitbasesPath,newEgbbCacheSize,egbb_load_type);
	g_printf("info string endgame bitbases loaded\n");
	fflush(stdout);
}

gboolean egbbProbe(const Position *position, gint* score) {
	gint piece[3], square[3];
	gint from, type, count;

	count = 0;
	piece[0] = piece[1] = piece[2] = _EMPTY;
	square[0] = square[1] = square[2] = _EMPTY;

	guint64 pieces = position->piece[White]|position->piece[Black];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		type = position->type[from];
		if(type != White_King && type != Black_King){
			square[count] = from;
			switch(type){
				case White_Pawn: 	{piece[count] = _WPAWN; break;}
				case White_Knight: 	{piece[count] = _WKNIGHT; break;}
				case White_Bishop: 	{piece[count] = _WBISHOP; break;}
				case White_Rook: 	{piece[count] = _WROOK; break;}
				case White_Queen: 	{piece[count] = _WQUEEN; break;}
				case Black_Pawn: 	{piece[count] = _BPAWN; break;}
				case Black_Knight: 	{piece[count] = _BKNIGHT; break;}
				case Black_Bishop: 	{piece[count] = _BBISHOP; break;}
				case Black_Rook: 	{piece[count] = _BROOK; break;}
				case Black_Queen: 	{piece[count] = _BQUEEN; break;}
			}
			count++;
		}
	}

	*score = probe_egbb(position->turn,
						getFirstBit64(position->piece[White_King]),
						getFirstBit64(position->piece[Black_King]),
						piece[0],square[0],piece[1],square[1],piece[2],square[2]);


	if(*score != _NOTFOUND){
		bbhits ++;
		return TRUE;
	}

    return FALSE;
}

gboolean egbbIsLoaded(){
	if(probe_egbb != NULL) return TRUE;
	return FALSE;
}

void egbbClose(){
	if(module != NULL) g_module_close(module);
}
