/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <stdlib.h>

#include "hash.h"
#include "threadSearch.h"
#include "moveDo.h"
#include "search.h"

/**
 * Transposition table
 */

guint hashSize = 64;
static guint hSize;
static guint hFree;
static guint16 hOld;
static HashData* data = NULL;

void hashCreate(void){
	guint dSize = sizeof(HashData);
	hFree = hSize = (hashSize * 1048576)/dSize;
	data = g_malloc(hSize*dSize);
	hashReset();
}

void hashDestroy(void){
	if(data != NULL) free(data);
	data = NULL;
}

guint hashUsage(void){
	return (guint)(((hSize-hFree)/(gfloat)hSize)*1000);
}

void hashWrite(HashData* hd, const guint64 key, const gint depth, const gint eval, const guint flag, 
				const Move move){
	guint64 hData;
	guint hdDepth;
	if(stopFlag == FALSE && parallelStopFlag == FALSE){
		hdDepth = hashDepth(hd->data);
		if(hdDepth <= depth || hashOld(hd->data) < hOld){
			if(hdDepth == 0 || hashOld(hd->data) < hOld) hFree --;
			hData = depth | (flag << 8) | (move << 16) | (((guint64)hOld) << 32) |  (((guint64)eval) << 48);
			hd->key = key ^ hData;
			hd->data = hData;
		}
	}
}

void hashWriteBestMove(const guint64 key, const gint depth, const gint eval, const Move move){
	HashData * hd = &data[key % hSize];
	guint64 hData;
	if(stopFlag == FALSE && parallelStopFlag == FALSE){
		guint hdDepth = hashDepth(hd->data);
		if(hdDepth == 0) hFree --;
		hData = depth | (EXACT << 8) | (move << 16) | (((guint64)hOld) << 32) | (((guint64)eval) << 48);
		hd->key = key ^ hData;
		hd->data = hData;
	}
}

gboolean hashFind(HashData ** hd, const guint64 key, guint64* hData){
	*hd = &data[key % hSize];
	*hData = (*hd)->data;
	guint64 hKey = (*hd)->key ^ (*hData);
	if(hKey == key){
		if(hashOld(*hData) < hOld){
			(*hd)->data &= ~(0xFFFFull<<32);
			(*hd)->data |= ((guint64)hOld)<<32;
			(*hd)->key = key ^ (*hd)->data;
			hFree --;
		}
		return TRUE;
	}
	return FALSE;
}

void hashIncrement(){
	gint i;
	guint64 key;
	if(hOld == G_MAXUINT16){
		for(i=0; i<hSize; i++){
			key = data[i].data ^ data[i].key;
			data[i].data &= ~(0xFFFFull<<32);
			data[i].key = key ^ data[i].data;
		}
		hOld = 1;
	}
	else hOld++;
	hFree = hSize;
}

void hashReset(void){
	gint i;
	for(i=0; i<hSize; i++) data[i].data = 0;
	hOld = 1;
	hFree = hSize;
}

void hashCollectPV(Position* position, Move pv[]){
	HashData* hd;
	guint64 hData;
	guint index = 0;
	while(hashFind(&hd,position->key,&hData)){
		if(index < Max_Pv-1 && hashMove(hData) != Empty){
			pv[index] = hashMove(hData);
			moveDo(position,pv[index]);
			index++;
		}
		else break;
	}
	
	pv[index] = Empty;
	while(index>0){
		moveUndo(position);
		index--;
	}
}

/**
 * Pawn hash
 */

guint pawnHashSize = 1;
static guint phSize;
static PawnHashData* pawnData = NULL;

void pawnHashCreate(void){  	 	 
       guint dSize = sizeof(PawnHashData); 		 
       phSize = (pawnHashSize * 1048576)/dSize; 		 
       pawnData = g_malloc(phSize*dSize); 		 
} 		 
  		 
void pawnHashDestroy(void){ 		 
       if(pawnData != NULL) free(pawnData); 		 
       pawnData = NULL; 		 
} 		 
void pawnHashWrite(const guint64 key, PawnHashData* pd, const gint eval[2], const guint64 passed[2]){
		guint64 evalNew = (guint32)eval[Opening] | (((guint64)eval[Endgame])<<32);
		pd->wPassedPawn = passed[White];
		pd->bPassedPawn = passed[Black];
		pd->eval = evalNew;
		pd->key = key ^ passed[White] ^ passed[Black] ^ evalNew;
}

gboolean pawnHashFind(const guint64 key, PawnHashData** pd, gint eval[2], guint64 passed[2]){	
		guint64 evalOld;
		(*pd) = &pawnData[key% phSize];
		passed[White] = (*pd)->wPassedPawn;
		passed[Black] = (*pd)->bPassedPawn;
		evalOld = (*pd)->eval;
		if(key == ((*pd)->key ^ passed[White] ^ passed[Black] ^ evalOld)){
			eval[Opening] = (gint)((gint32)(evalOld & 0xFFFFFFFF));
			eval[Endgame] = (gint)((gint32)(evalOld >> 32));
			return TRUE;
		}
		return FALSE; 		 
}
