/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef HASH_H
#define HASH_H

#include <glib.h>
#include "move.h"
#include "position.h"

#define EXACT 0
#define ALPHA 1
#define BETA 2

#define hashDepth(data)	((guint)(data & 0xFF))
#define hashFlag(data)	((guint)(data>>8 & 0xFF))
#define hashMove(data)	((guint)(data>>16 & 0xFFFF))
#define hashOld(data)	((guint)(data>>32) & 0xFFFF)
#define hashEval(data)	((gint16)(data>>48))

/**
 * hash data structure:
 * 00000000 00000000 00000000 00000000 00000000 00000000 00000000 11111111 - depth
 * 00000000 00000000 00000000 00000000 00000000 00000000 11111111 00000000 - flag
 * 00000000 00000000 00000000 00000000 11111111 11111111 00000000 00000000 - move
 * 00000000 00000000 11111111 11111111 00000000 00000000 00000000 00000000 - old
 * 11111111 11111111 00000000 00000000 00000000 00000000 00000000 00000000 - eval
 */

struct hashData{
	guint64 key;
	guint64 data;
};
typedef struct hashData HashData;

extern guint hashSize;
void hashCreate(void);
void hashDestroy(void);
guint hashUsage(void);
void hashReset();
void hashIncrement();
gboolean hashFind(HashData** hd, const guint64 key, guint64* data);
void hashWrite(HashData* hd, const guint64 key, const gint depth, const gint eval, const guint flag, 
				const Move move);
void hashCollectPV(Position* position, Move pv[]);
void hashWriteBestMove(const guint64 key, const gint depth, const gint eval, const Move move);

struct pHashData{  	 	 
       guint64 key;
	   guint64 eval; /* first 32 bits for opening game evaluation and last 32 bits for endgame evaluation */
       guint64 wPassedPawn; 		 
       guint64 bPassedPawn; 		 
}; 		 
typedef struct pHashData PawnHashData;

void pawnHashCreate(void);  	 	 
void pawnHashDestroy(void); 		 
void pawnHashWrite(const guint64 key, PawnHashData* pd, const gint32 eval[2], const guint64 passed[2]);
gboolean pawnHashFind(const guint64 key, PawnHashData** pd, gint32 eval[2], guint64 passed[2]);

#endif
