/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic                                  *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <glib/gprintf.h>

#include "moveDo.h"
#include "moveGenerator.h"
#include "bit.h"

gboolean moveDo(Position* position, const Move move);
inline void normalMoveDo(Position* position, const Move move);
inline void normalMoveUndo(Position* position, const Move move);
inline void captureMoveDo(Position* position, const Move move);
inline void captureMoveUndo(Position* position, const Move move);
inline void enPassantMoveDo(Position* position, const Move move);
inline void enPassantMoveUndo(Position* position, const Move move);
inline gboolean castleMoveDo(Position* position, const Move move);
inline void castleMoveUndo(Position* position, const Move move);
inline void promoteMoveDo(Position* position, const Move move, const guint promote);
inline void promoteMoveUndo(Position* position, const Move move, const guint promote);
inline void nullMoveDo(Position* position);
inline gboolean positionGood(const Position* position);
inline gboolean whiteCastleKingSide(const Position* position);
inline gboolean whiteCastleQueenSide(const Position* position);
inline gboolean blackCastleKingSide(const Position* position);
inline gboolean blackCastleQueenSide(const Position* position);

Move nextMoveDo(Position* position, Moves* moves){
	guint index;
	gint value;
	Move move;
	if(moves->state == START){
		dinamicMovesGenerate(position, moves);
		moves->state = DINAMIC;
	}
	while(moves->next < moves->size){
		if(moves->next >= moves->toSort){
			moves->toSort = moves->size;
			for(index = moves->size-1; index>moves->next; index--){
				if(moves->eval[index-1] < moves->eval[index]){
					value = moves->eval[index];
					moves->eval[index] = moves->eval[index-1];
					moves->eval[index-1] = value;
			
					move = moves->move[index];
					moves->move[index] = moves->move[index-1];
					moves->move[index-1] = move;
					moves->toSort = index;
				}
			}
		}
		if(moveDo(position,moves->move[moves->next])){
			moves->next ++;
			return moves->move[moves->next-1];
		}
		moves->next ++;
	}
	if(moves->state == DINAMIC){
		staticMovesGenerate(position, moves);
		moves->state = STATIC;
		while(moves->next < moves->size){
			if(moves->next >= moves->toSort){
				moves->toSort = moves->size;
				for(index = moves->size-1; index>moves->next; index--){
					if(moves->eval[index-1] < moves->eval[index]){
						value = moves->eval[index];
						moves->eval[index] = moves->eval[index-1];
						moves->eval[index-1] = value;
			
						move = moves->move[index];
						moves->move[index] = moves->move[index-1];
						moves->move[index-1] = move;
						moves->toSort = index;
					}
				}
			}
			if(moveDo(position,moves->move[moves->next])){
				moves->next ++;
				return moves->move[moves->next-1];
			}
			moves->next ++;
		}		
	}
	return Empty;
}

Move nextDinamicMoveDo(Position* position, Moves* moves){
	guint index;
	gint value;
	Move move;
	if(moves->state == START){
		dinamicMovesGenerate(position, moves);
		moves->state = DINAMIC;
	}
	while(moves->next < moves->size){
		if(moves->next >= moves->toSort){
			moves->toSort = moves->size;
			for(index = moves->size-1; index>moves->next; index--){
				if(moves->eval[index-1] < moves->eval[index]){
					value = moves->eval[index];
					moves->eval[index] = moves->eval[index-1];
					moves->eval[index-1] = value;
			
					move = moves->move[index];
					moves->move[index] = moves->move[index-1];
					moves->move[index-1] = move;
					moves->toSort = index;
				}
			}
		}
		if(moveDo(position,moves->move[moves->next])){
			moves->next ++;
			return moves->move[moves->next-1];
		}
		moves->next ++;
	}
	return Empty;
}

gboolean moveDo(Position* position, const Move move){
	History* history = &position->history[position->ply];
	history->move = move;
	history->key = position->key;
	history->capture = position->capture;
	history->staticMoves = position->staticMoves;
	history->enPassant = position->enPassant;
	history->castle = position->castle;
	history->inCheck = position->inCheck;
	history->material[White] = position->material[White];
	history->material[Black] = position->material[Black];
	history->pvt[Opening] = position->pvt[Opening];
	history->pvt[Endgame] = position->pvt[Endgame];
	history->extend = position->extend;
	history->phase = position->phase;

	#ifndef G_DISABLE_ASSERT 
	if(move != Null_Move && position->type[moveGetFrom(move)]%2 != position->turn){
		positionPrint(position);
		g_printf("Wrong move (");
		movePrint(move);
		g_printf(")\n");
		g_error("!!\n");
	}
	#endif
	if(position->enPassant != 0){
		position->key ^= enPassantKey[file(position->enPassant)];
		position->enPassant = 0;
	}
	position->capture = Empty;
	position->extend = FALSE;
	position->staticMoves ++;

	switch(moveGetType(move)){
		case Normal: 	{ normalMoveDo(position, move); break;}
		case Capture:	{ captureMoveDo(position, move); break;}
		case En_Passant:{ enPassantMoveDo(position, move); break;}
		case Castle:	{
			if(!castleMoveDo(position,move)){
				position->enPassant = history->enPassant;
				position->capture = history->capture;
				position->extend = history->extend;
				position->staticMoves--;
				#ifndef G_DISABLE_ASSERT 
				positionIsOk(position);
				#endif
				return FALSE;
			}else{
				position->turn ^= 1; 
				position->key ^= turnKey;
				position->ply++;
				position->inCheck = positionIsInCheck(position); /* Speed up (rook only)*/
				if(position->inCheck) position->extend = TRUE;
				#ifndef G_DISABLE_ASSERT 
				positionIsOk(position);
				#endif
				return TRUE;
			}
		}
		case Promote_Queen:	{ promoteMoveDo(position,move,Queen); break;}
		case Promote_Rook: 	{ promoteMoveDo(position,move,Rook); break;}
		case Promote_Bishop:	{ promoteMoveDo(position,move,Bishop); break;}
		case Promote_Knight:	{ promoteMoveDo(position,move,Knight); break;}
		case Null_Move:	{
			position->turn ^= 1; 
			position->key ^= turnKey;
			position->ply++;
			position->staticMoves = 0;
			#ifndef G_DISABLE_ASSERT 
			positionIsOk(position);
			#endif
			return TRUE;
		}
		default: { g_error("Wrong move type!\n"); }
	};
	position->turn ^= 1; 
	position->key ^= turnKey;
	position->ply++;
	if(!positionGood(position)){
		moveUndo(position);
		return FALSE;
	}
	position->inCheck = positionIsInCheck(position); /* Speed up */
	if(position->inCheck) position->extend = TRUE;
	#ifndef G_DISABLE_ASSERT 
	positionIsOk(position);
	#endif
	return TRUE;
}


void moveUndo(Position* position){
	position->ply --;
	History* history = &position->history[position->ply];
	Move move = history->move;
	switch(moveGetType(move)){
		case Normal: 	{ normalMoveUndo(position, move); break;}
		case Capture:	{ captureMoveUndo(position, move); break;}
		case En_Passant:{ enPassantMoveUndo(position, move); break;}
		case Castle:	{ castleMoveUndo(position,move); break; }
		case Promote_Queen:	{ promoteMoveUndo(position,move,Queen); break;}
		case Promote_Rook: 	{ promoteMoveUndo(position,move,Rook); break;}
		case Promote_Bishop:{ promoteMoveUndo(position,move,Bishop); break;}
		case Promote_Knight:{ promoteMoveUndo(position,move,Knight); break;}
		case Null_Move:	{ break;}
		default: {g_error("\nWrong move type!\n");}
	};
	position->key = history->key;
	position->staticMoves = history->staticMoves;
	position->enPassant = history->enPassant;
	position->castle = history->castle;
	position->inCheck = history->inCheck;
	position->material[White] = history->material[White];
	position->material[Black] = history->material[Black];
	position->pvt[Opening] = history->pvt[Opening];
	position->pvt[Endgame] = history->pvt[Endgame];
	position->capture = history->capture;
	position->extend = history->extend;
	position->phase = history->phase;
	position->turn ^= 1;	
	#ifndef G_DISABLE_ASSERT 
	positionIsOk(position);
	#endif
}

inline void normalMoveDo(Position* position, const Move move){
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	guint piece = position->type[from];
	guint color = piece % 2;
	guint64 bitMove = 1ull << from | 1ull << to;
	position->piece[color] ^= bitMove;
	position->piece[piece] ^= bitMove;
	position->rotateL90 ^= 1ull << rotateL90[from] | 1ull << rotateL90[to]; 
	position->rotateL45 ^= 1ull << rotateL45[from] | 1ull << rotateL45[to];
	position->rotateR45 ^= 1ull << rotateR45[from] | 1ull << rotateR45[to];
	position->key ^= squareKey[piece][from] ^ squareKey[piece][to];
	position->type[from] = Empty;
	position->type[to] = piece;
	position->pvt[Opening] += pvt[piece][to][Opening] - pvt[piece][from][Opening];
	position->pvt[Endgame] += pvt[piece][to][Endgame] - pvt[piece][from][Endgame];
	if(color == White){
		if(piece == White_Pawn){
			position->staticMoves = 0;
			if(to - from == 16 && (position->piece[Black_Pawn] & ((1ull<<(to+1)) | (1ull<<(to-1))))){
				position->enPassant = to - 8;
				position->key ^= enPassantKey[file(position->enPassant)];
			}
			if(to >= a7) position->extend = TRUE;
		}
		if(position->castle & WHITE_CASTLE){
			if(position->castle & WHITE_KING_CASTLE){
				if(from == e1 || from == h1){
					position->key ^= castleKey[White_King_Castle];
					position->castle &= ~WHITE_KING_CASTLE;
				}
			}
			if(position->castle & WHITE_QUEEN_CASTLE){
				if(from == e1 || from == a1){
					position->key ^= castleKey[White_Queen_Castle];
					position->castle &= ~WHITE_QUEEN_CASTLE;
				}
			}
		}
	}
	else{
		if(piece == Black_Pawn){
			position->staticMoves = 0;
			if(from - to == 16 && (position->piece[White_Pawn] & ((1ull<<(to+1)) | (1ull<<(to-1))))){
				position->enPassant = to + 8;
				position->key ^= enPassantKey[file(position->enPassant)];
			}
			if(to <= h2) position->extend = TRUE;
		}
		if(position->castle & BLACK_CASTLE){
			if(position->castle & BLACK_KING_CASTLE){
				if(from == e8 || from == h8){
					position->key ^= castleKey[Black_King_Castle];
					position->castle &= ~BLACK_KING_CASTLE;
				}
			}
			if(position->castle & BLACK_QUEEN_CASTLE){
				if(from == e8 || from == a8){
					position->key ^= castleKey[Black_Queen_Castle];
					position->castle &= ~BLACK_QUEEN_CASTLE;
				}
			}
		}
	}
}

inline void normalMoveUndo(Position* position, const Move move){
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	guint piece = position->type[to];
	guint color = piece % 2;
	guint64 bitMove = 1ull << from | 1ull << to;
	position->piece[color] ^= bitMove;
	position->piece[piece] ^= bitMove;
	position->rotateL90 ^= 1ull << rotateL90[from] | 1ull << rotateL90[to]; 
	position->rotateL45 ^= 1ull << rotateL45[from] | 1ull << rotateL45[to];
	position->rotateR45 ^= 1ull << rotateR45[from] | 1ull << rotateR45[to];
	position->type[from] = piece;
	position->type[to] = Empty;
}

inline void captureMoveDo(Position* position, const Move move){
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	guint piece = position->type[from];
	guint capture = position->type[to];
	guint color = piece % 2;
	guint64 bitTo = 1ull << to;
	guint64 bitMove = 1ull<<from | bitTo;
	position->piece[color] ^= bitMove;
	position->piece[piece] ^= bitMove;
	position->piece[color^1] ^= bitTo;
	position->piece[capture] ^= bitTo;
	position->rotateL90 ^= 1ull << rotateL90[from]; 
	position->rotateL45 ^= 1ull << rotateL45[from];
	position->rotateR45 ^= 1ull << rotateR45[from];
	position->key ^= squareKey[piece][from] ^ squareKey[piece][to] ^ squareKey[capture][to];
	position->type[from] = Empty;
	position->type[to] = piece;
	position->pvt[Opening] += pvt[piece][to][Opening] - pvt[piece][from][Opening] - pvt[capture][to][Opening];
	position->pvt[Endgame] += pvt[piece][to][Endgame] - pvt[piece][from][Endgame] - pvt[capture][to][Endgame];
	position->material[color^1] -= pieceValue[capture];
	position->staticMoves = 0;
	position->capture = capture;
	position->phase -= piecePhase[capture];
	if(color == White){
		if(piece == White_Pawn && to >= a7) position->extend = TRUE;
		if(position->castle & WHITE_CASTLE){
			if(position->castle & WHITE_KING_CASTLE){
				if(from == e1 || from == h1){
					position->key ^= castleKey[White_King_Castle];
					position->castle &= ~WHITE_KING_CASTLE;
				}
			}
			if(position->castle & WHITE_QUEEN_CASTLE){
				if(from == e1 || from == a1){
					position->key ^= castleKey[White_Queen_Castle];
					position->castle &= ~WHITE_QUEEN_CASTLE;
				}
			}
		}
		if(position->castle & BLACK_CASTLE){
			if(position->castle & BLACK_KING_CASTLE){
				if(to == h8){
					position->key ^= castleKey[Black_King_Castle];
					position->castle &= ~BLACK_KING_CASTLE;
				}
			}
			if(position->castle & BLACK_QUEEN_CASTLE){
				if(to == a8){
					position->key ^= castleKey[Black_Queen_Castle];
					position->castle &= ~BLACK_QUEEN_CASTLE;
				}
			}
		}
	}
	else{
		if(piece == Black_Pawn && to <= h2) position->extend = TRUE;
		if(position->castle & BLACK_CASTLE){
			if(position->castle & BLACK_KING_CASTLE){
				if(from == e8 || from == h8){
					position->key ^= castleKey[Black_King_Castle];
					position->castle &= ~BLACK_KING_CASTLE;
				}
			}
			if(position->castle & BLACK_QUEEN_CASTLE){
				if(from == e8 || from == a8){
					position->key ^= castleKey[Black_Queen_Castle];
					position->castle &= ~BLACK_QUEEN_CASTLE;
				}
			}
		}
		if(position->castle & WHITE_CASTLE){
			if(position->castle & WHITE_KING_CASTLE){
				if(to == h1){
					position->key ^= castleKey[White_King_Castle];
					position->castle &= ~WHITE_KING_CASTLE;
				}
			}
			if(position->castle & WHITE_QUEEN_CASTLE){
				if(to == a1){
					position->key ^= castleKey[White_Queen_Castle];
					position->castle &= ~WHITE_QUEEN_CASTLE;
				}
			}
		}
	}
}

inline void captureMoveUndo(Position* position, const Move move){
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	guint piece = position->type[to];
	guint capture = position->capture;
	guint color = piece % 2;
	guint64 bitTo = 1ull << to;
	guint64 bitMove = 1ull<<from | bitTo;
	position->piece[color] ^= bitMove;
	position->piece[piece] ^= bitMove;
	position->piece[color^1] ^= bitTo;
	position->piece[capture] ^= bitTo;
	position->rotateL90 ^= 1ull << rotateL90[from]; 
	position->rotateL45 ^= 1ull << rotateL45[from];
	position->rotateR45 ^= 1ull << rotateR45[from];
	position->type[from] = piece;
	position->type[to] = capture;
}

inline void enPassantMoveDo(Position* position, const Move move){
	guint piece, capture, cSquare, color;
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	if(to < a4){
		color = Black;
		piece = Black_Pawn;
		capture = White_Pawn;
		cSquare = to+8;
	}
	else{
		color = White;
		piece = White_Pawn;
		capture = Black_Pawn;
		cSquare = to-8;
	}
	guint64 bitMove = 1ull<<from | 1ull<<to;
	guint64 bitCapture = 1ull<<cSquare;
	position->piece[color] ^= bitMove;
	position->piece[piece] ^= bitMove;
	position->piece[color^1] ^= bitCapture;
	position->piece[capture] ^= bitCapture;
	position->rotateL90 ^= 1ull << rotateL90[from] | 1ull << rotateL90[to] | 1ull << rotateL90[cSquare]; 
	position->rotateL45 ^= 1ull << rotateL45[from] | 1ull << rotateL45[to] | 1ull << rotateL45[cSquare];
	position->rotateR45 ^= 1ull << rotateR45[from] | 1ull << rotateR45[to] | 1ull << rotateR45[cSquare];
	position->key ^= squareKey[piece][from] ^ squareKey[piece][to] ^ squareKey[capture][cSquare];
	position->type[from] = Empty;
	position->type[to] = piece;
	position->type[cSquare] = Empty;
	position->pvt[Opening] += pvt[piece][to][Opening]-pvt[piece][from][Opening]-pvt[capture][cSquare][Opening];
	position->pvt[Endgame] += pvt[piece][to][Endgame]-pvt[piece][from][Endgame]-pvt[capture][cSquare][Endgame];
	position->material[color^1] -= Pawn_Value;
	position->staticMoves = 0;
}

inline void enPassantMoveUndo(Position* position, const Move move){
	guint piece, capture, cSquare, color;
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	if(to < a4){
		color = Black;
		piece = Black_Pawn;
		capture = White_Pawn;
		cSquare = to+8;
	}
	else{
		color = White;
		piece = White_Pawn;
		capture = Black_Pawn;
		cSquare = to-8;
	}
	guint64 bitMove = 1ull<<from | 1ull<<to;
	guint64 bitCapture = 1ull<<cSquare;
	position->piece[color] ^= bitMove;
	position->piece[piece] ^= bitMove;
	position->piece[color^1] ^= bitCapture;
	position->piece[capture] ^= bitCapture;
	position->rotateL90 ^= 1ull << rotateL90[from] | 1ull << rotateL90[to] | 1ull << rotateL90[cSquare]; 
	position->rotateL45 ^= 1ull << rotateL45[from] | 1ull << rotateL45[to] | 1ull << rotateL45[cSquare];
	position->rotateR45 ^= 1ull << rotateR45[from] | 1ull << rotateR45[to] | 1ull << rotateR45[cSquare];
	position->type[from] = piece;
	position->type[to] = Empty;
	position->type[cSquare] = capture;
}

inline gboolean castleMoveDo(Position* position, const Move move){
	guint piece, rook, rFrom, rTo, color;
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	if(to < a4){
		if(to > e1){
			if(!whiteCastleKingSide(position)) return FALSE;
			rFrom = h1;
			rTo = f1;
		}
		else{
			if(!whiteCastleQueenSide(position)) return FALSE;
			rFrom = a1;
			rTo = d1;
		}
		if(position->castle & WHITE_KING_CASTLE){
			position->key^= castleKey[White_King_Castle];
			position->castle &= ~WHITE_KING_CASTLE;
		}
		if(position->castle & WHITE_QUEEN_CASTLE){
			position->key^= castleKey[White_Queen_Castle];
			position->castle &= ~WHITE_QUEEN_CASTLE;
		}
		color = White;
		piece = White_King;
		rook = White_Rook;
	}
	else{
		if(to > e8){
			if(!blackCastleKingSide(position)) return FALSE;
			rFrom = h8;
			rTo = f8;
		}
		else{
			if(!blackCastleQueenSide(position)) return FALSE;
			rFrom = a8;
			rTo = d8;
		}
		if(position->castle & BLACK_KING_CASTLE){
			position->key^= castleKey[Black_King_Castle];
			position->castle &= ~BLACK_KING_CASTLE;
		}
		if(position->castle & BLACK_QUEEN_CASTLE){
			position->key^= castleKey[Black_Queen_Castle];
			position->castle &= ~BLACK_QUEEN_CASTLE;
		}
		color = Black;
		piece = Black_King;
		rook = Black_Rook;
	}
	guint64 bitMove = 1ull<<from | 1ull<<to;
	guint64 bitRook = 1ull<<rFrom | 1ull<<rTo;
	position->piece[color] ^= bitMove | bitRook;
	position->piece[piece] ^= bitMove;
	position->piece[rook] ^= bitRook;
	position->rotateL90^=1ull<<rotateL90[from]|1ull<<rotateL90[to]|1ull<<rotateL90[rFrom]|1ull<<rotateL90[rTo]; 
	position->rotateL45^=1ull<<rotateL45[from]|1ull<<rotateL45[to]|1ull<<rotateL45[rFrom]|1ull<<rotateL45[rTo];
	position->rotateR45^=1ull<<rotateR45[from]|1ull<<rotateR45[to]|1ull<<rotateR45[rFrom]|1ull<<rotateR45[rTo];
	position->key^=squareKey[piece][from]^squareKey[piece][to]^squareKey[rook][rFrom]^squareKey[rook][rTo];
	position->type[from] = Empty;
	position->type[to] = piece;
	position->type[rFrom] = Empty;
	position->type[rTo] = rook;
	position->pvt[Opening] += pvt[piece][to][Opening]-pvt[piece][from][Opening] +
								pvt[rook][rTo][Opening]-pvt[rook][rFrom][Opening];
	position->pvt[Endgame] += pvt[piece][to][Endgame]-pvt[piece][from][Endgame] +
								pvt[rook][rTo][Endgame]-pvt[rook][rFrom][Endgame];
	position->staticMoves = 0;
	return TRUE;
}

inline void castleMoveUndo(Position* position, const Move move){
	guint piece, rook, rFrom, rTo, color;
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	if(to < a4){
		color = White;
		piece = White_King;
		rook = White_Rook;
		if(to > e1){
			rFrom = h1;
			rTo = f1;
		}
		else{
			rFrom = a1;
			rTo = d1;
		}
	}
	else{
		color = Black;
		piece = Black_King;
		rook = Black_Rook;
		if(to > e8){
			rFrom = h8;
			rTo = f8;
		}
		else{
			rFrom = a8;
			rTo = d8;
		}
	}
	guint64 bitMove = 1ull<<from | 1ull<<to;
	guint64 bitRook = 1ull<<rFrom | 1ull<<rTo;
	position->piece[color] ^= bitMove | bitRook;
	position->piece[piece] ^= bitMove;
	position->piece[rook] ^= bitRook;
	position->rotateL90^=1ull<<rotateL90[from]|1ull<<rotateL90[to]|1ull<<rotateL90[rFrom]|1ull<<rotateL90[rTo]; 
	position->rotateL45^=1ull<<rotateL45[from]|1ull<<rotateL45[to]|1ull<<rotateL45[rFrom]|1ull<<rotateL45[rTo];
	position->rotateR45^=1ull<<rotateR45[from]|1ull<<rotateR45[to]|1ull<<rotateR45[rFrom]|1ull<<rotateR45[rTo];
	position->type[from] = piece;
	position->type[to] = Empty;
	position->type[rFrom] = rook;
	position->type[rTo] = Empty;
}

inline void promoteMoveDo(Position* position, const Move move, const guint promote){
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	guint piece, pPiece, capture, color;
	capture = position->type[to];
	if(to>a4){
		color = White;
		piece = White_Pawn;
		pPiece = promote + White; 	
	}
	else{
		color = Black;
		piece = Black_Pawn;
		pPiece = promote + Black; 	
	}
	guint64 bitFrom = 1ull<<from;
	guint64 bitTo = 1ull<<to;
	position->piece[color] ^= bitFrom | bitTo;
	position->piece[piece] ^= bitFrom;
	position->piece[pPiece] ^= bitTo;
	position->rotateL90 ^= 1ull << rotateL90[from] | 1ull << rotateL90[to]; 
	position->rotateL45 ^= 1ull << rotateL45[from] | 1ull << rotateL45[to];
	position->rotateR45 ^= 1ull << rotateR45[from] | 1ull << rotateR45[to];
	position->key ^= squareKey[piece][from] ^ squareKey[pPiece][to];
	position->type[from] = Empty;
	position->type[to] = pPiece;
	position->pvt[Opening] += pvt[pPiece][to][Opening] - pvt[piece][from][Opening];
	position->pvt[Endgame] += pvt[pPiece][to][Endgame] - pvt[piece][from][Endgame];
	position->material[color] += pieceValue[pPiece] - Pawn_Value;
	position->staticMoves = 0;
	position->phase += piecePhase[promote];
	if(capture != Empty){
		if(color == White){
			if(position->castle & BLACK_CASTLE){
				if(position->castle & BLACK_KING_CASTLE){
					if(to == h8){
						position->key ^= castleKey[Black_King_Castle];
						position->castle &= ~BLACK_KING_CASTLE;
					}
				}
				if(position->castle & BLACK_QUEEN_CASTLE){
					if(to == a8){
						position->key ^= castleKey[Black_Queen_Castle];
						position->castle &= ~BLACK_QUEEN_CASTLE;
					}
				}
			}
		}
		else{
			if(position->castle & WHITE_CASTLE){
				if(position->castle & WHITE_KING_CASTLE){
					if(to == h1){
						position->key ^= castleKey[White_King_Castle];
						position->castle &= ~WHITE_KING_CASTLE;
					}
				}
				if(position->castle & WHITE_QUEEN_CASTLE){
					if(to == a1){
						position->key ^= castleKey[White_Queen_Castle];
						position->castle &= ~WHITE_QUEEN_CASTLE;
					}
				}
			}
		}
		position->capture = capture;
		position->piece[color^1] ^= bitTo;
		position->piece[capture] ^= bitTo;
		position->rotateL90 ^= 1ull << rotateL90[to]; 
		position->rotateL45 ^= 1ull << rotateL45[to];
		position->rotateR45 ^= 1ull << rotateR45[to];
		position->key ^= squareKey[capture][to];
		position->pvt[Opening] -= pvt[capture][to][Opening];
		position->pvt[Endgame] -= pvt[capture][to][Endgame];
		position->material[color^1] -= pieceValue[capture];
		position->extend = TRUE;
	}
}

inline void promoteMoveUndo(Position* position, const Move move, const guint promote){
	guint from = moveGetFrom(move);
	guint to = moveGetTo(move);
	guint piece, pPiece, capture, color;
	if(to>a4){
		color = White;
		piece = White_Pawn;
		pPiece = promote + White; 	
	}
	else{
		color = Black;
		piece = Black_Pawn;
		pPiece = promote + Black; 	
	}
	guint64 bitFrom = 1ull<<from;
	guint64 bitTo = 1ull<<to;
	position->piece[color] ^= bitFrom | bitTo;
	position->piece[piece] ^= bitFrom;
	position->piece[pPiece] ^= bitTo;
	position->rotateL90 ^= 1ull << rotateL90[from] | 1ull << rotateL90[to]; 
	position->rotateL45 ^= 1ull << rotateL45[from] | 1ull << rotateL45[to];
	position->rotateR45 ^= 1ull << rotateR45[from] | 1ull << rotateR45[to];
	position->type[from] = piece;
	position->type[to] = Empty;
	capture = position->capture;
	if(capture != Empty){
		position->type[to] = capture;
		position->piece[color^1] ^= bitTo;
		position->piece[capture] ^= bitTo;
		position->rotateL90 ^= 1ull << rotateL90[to]; 
		position->rotateL45 ^= 1ull << rotateL45[to];
		position->rotateR45 ^= 1ull << rotateR45[to];
	}
}

inline gboolean positionGood(const Position* position){
	guint player = position->turn;
	guint opponent = player ^ 1;
	guint square = getFirstBit64(position->piece[King+opponent]);
	if(positionAttackRook(position,square)&(position->piece[Rook+player]|position->piece[Queen+player]))
		return FALSE;
	if(positionAttackBishop(position,square)&(position->piece[Bishop+player]|position->piece[Queen+player]))
		return FALSE;
	if(attackKnight[square]&position->piece[Knight+player]) return FALSE;
	if(attackPawn[opponent][square]&position->piece[Pawn+player]) return FALSE;
	if(attackKing[square]&position->piece[King+player]) return FALSE;
	return TRUE;
}

inline gboolean whiteCastleKingSide(const Position* position){
	if((positionAttackRook(position,e1)|positionAttackRook(position,f1)|positionAttackRook(position,g1) )&
		(position->piece[Black_Rook]|position->piece[Black_Queen])) return FALSE;
	if((positionAttackBishop(position,e1)|positionAttackBishop(position,f1)|positionAttackBishop(position,g1))&
		(position->piece[Black_Bishop]|position->piece[Black_Queen])) return FALSE;
	if(16309248ull&position->piece[Black_Knight]) return FALSE;
	if(63488ull&position->piece[Black_Pawn]) return FALSE;
	if(63736ull&position->piece[Black_King]) return FALSE;
	return TRUE;
}

inline gboolean whiteCastleQueenSide(const Position* position){
	if((positionAttackRook(position,c1)|positionAttackRook(position,d1)|positionAttackRook(position,e1) )&
		(position->piece[Black_Rook]|position->piece[Black_Queen])) return FALSE;
	if((positionAttackBishop(position,c1)|positionAttackBishop(position,d1)|positionAttackBishop(position,e1))&
		(position->piece[Black_Bishop]|position->piece[Black_Queen])) return FALSE;
	if(4093696ull&position->piece[Black_Knight]) return FALSE;
	if(15872ull&position->piece[Black_Pawn]) return FALSE;
	if(15934ull&position->piece[Black_King]) return FALSE;
	return TRUE;
}

inline gboolean blackCastleKingSide(const Position* position){
	if((positionAttackRook(position,e8)|positionAttackRook(position,f8)|positionAttackRook(position,g8))&
		(position->piece[White_Rook]|position->piece[White_Queen])) return FALSE;
	if((positionAttackBishop(position,e8)|positionAttackBishop(position,f8)|positionAttackBishop(position,g8))&
		(position->piece[White_Bishop]|position->piece[White_Queen])) return FALSE;
	if(62197173760032768ull&position->piece[White_Knight]) return FALSE;
	if(69805794224242688ull&position->piece[White_Pawn]) return FALSE;
	if(17940089115630370816ull&position->piece[White_King]) return FALSE;
	return TRUE;
}

inline gboolean blackCastleQueenSide(const Position* position){
	if((positionAttackRook(position,c8)|positionAttackRook(position,d8)|positionAttackRook(position,e8))&
		(position->piece[White_Rook]|position->piece[White_Queen])) return FALSE;
	if( (positionAttackBishop(position,c8)|positionAttackBishop(position,d8)|positionAttackBishop(position,e8))&
		(position->piece[White_Bishop]|position->piece[White_Queen]) ) return FALSE;
	if(33563691949490176ull&position->piece[White_Knight]) return FALSE;
	if(17451448556060672ull&position->piece[White_Pawn]) return FALSE;
	if(4485022278907592704ull&position->piece[White_King]) return FALSE;
	return TRUE;
}

void stringMovesDo(Position* position, gchar moves[]){
	guint index=0, gindex;
	RootMoves gmoves;
	gchar smove[6], sgmove[6];
	gboolean found;
	gboolean last = FALSE;
	while(!last){
		rootMovesGenerate(position, &gmoves);
		smove[0] = '\0';
		sscanf(moves,"%s",smove);
		if(moves[strlen(smove)] == '\0' || moves[strlen(smove)] == '\n') last = TRUE;
		if(strlen(smove) == 0) break;
		moves = g_strstr_len(moves, strlen(moves), smove) + strlen(smove);
		found = FALSE;
		for(gindex = 0; gindex<gmoves.size; gindex++){
			moveToString(gmoves.move[gindex].move, sgmove);
			if(!g_ascii_strcasecmp(smove, sgmove)){
				found = TRUE;
				if(!moveDo(position,gmoves.move[gindex].move)){
					positionPrint(position);
					g_error("Wrong move:%s\n",smove);
				}
				break;
			}		
		}
		if(!found){
			positionPrint(position);
		    g_error("Can not make move (%s)!\n",smove);
		}
		index++;
	}
}

