/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "moveGenerator.h"
#include "moveDo.h"
#include "bit.h"
#include "move.h"
#include "search.h"

void dinamicMovesGenerate(const Position* position, Moves* moves){
	guint64 pieces, dest, attack, free;
	guint from, to, promote, player, opponent;

	player = position->turn;
	opponent = player ^ 1;
	free = ~(position->piece[White] | position->piece[Black]);
	moves->size = 0;
	
	if(player == White){
		pieces = position->piece[White_Pawn];
		dest = shiftUp(pieces) & free & RANK_8;
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to - 8;
			for(promote = Promote_Queen; promote >= Promote_Knight; promote-=2){
					moves->move[moves->size] = moveGenerate(promote,from,to);
					moves->eval[moves->size] = pieceValue[promote];
					moves->size ++;
			}
		}
		dest = shiftUpRight(pieces) & position->piece[Black];
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to - 9;
			if(to >= a8){
				for(promote = Promote_Queen; promote >= Promote_Knight; promote-=2){
					moves->move[moves->size] = moveGenerate(promote,from,to);
					moves->eval[moves->size] = pieceValue[promote] + pieceValue[position->type[to]];
					moves->size ++;
				}
			}
			else{
				moves->move[moves->size] = moveGenerate(Capture,from,to);
				moves->eval[moves->size] = pieceValue[position->type[to]]-Pawn_Value;
				moves->size ++;
			}
		}
		dest = shiftUpLeft(pieces) & position->piece[Black];
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to - 7;
			if(to >= a8){
				for(promote = Promote_Queen; promote >= Promote_Knight; promote-=2){
					moves->move[moves->size] = moveGenerate(promote,from,to);
					moves->eval[moves->size] = pieceValue[promote]+pieceValue[position->type[to]];
					moves->size ++;
				}
			}
			else{
				moves->move[moves->size] = moveGenerate(Capture,from,to);
				moves->eval[moves->size] = pieceValue[position->type[to]]-Pawn_Value;
				moves->size ++;
			}
		}
		if(position->enPassant){
			dest = pieces & attackPawn[Black][position->enPassant];
			while(dest){
				from = getFirstBitAndClear64(&dest);
				moves->move[moves->size] = moveGenerate(En_Passant,from,position->enPassant);
				moves->eval[moves->size] = 0;
				moves->size ++;
			}
		}
	}else{
		pieces = position->piece[Black_Pawn];
		dest = shiftDown(pieces) & free & RANK_1;
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to + 8;
			for(promote = Promote_Queen; promote >= Promote_Knight; promote-=2){
				moves->move[moves->size] = moveGenerate(promote,from,to);
				moves->eval[moves->size] = pieceValue[promote];
				moves->size ++;
			}
		}
		dest = shiftDownRight(pieces) & position->piece[White];
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to + 7;
			if(to <= h1){
				for(promote = Promote_Queen; promote >= Promote_Knight; promote-=2){
					moves->move[moves->size] = moveGenerate(promote,from,to);
					moves->eval[moves->size] = pieceValue[promote] + pieceValue[position->type[to]];
					moves->size ++;
				}
			}
			else{
				moves->move[moves->size] = moveGenerate(Capture,from,to);
				moves->eval[moves->size] = pieceValue[position->type[to]] - Pawn_Value;
				moves->size ++;
			}
		}
		dest = shiftDownLeft(pieces) & position->piece[White];
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to + 9;
			if(to <= h1){
				for(promote = Promote_Queen; promote >= Promote_Knight; promote-=2){
					moves->move[moves->size] = moveGenerate(promote,from,to);
					moves->eval[moves->size] = pieceValue[promote] + pieceValue[position->type[to]];
					moves->size ++;
				}
			}
			else{
				moves->move[moves->size] = moveGenerate(Capture,from,to);
				moves->eval[moves->size] = pieceValue[position->type[to]] - Pawn_Value;
				moves->size ++;
			}
		}
		if(position->enPassant){
			dest = pieces & attackPawn[White][position->enPassant];
			while(dest){
				from = getFirstBitAndClear64(&dest);
				moves->move[moves->size] = moveGenerate(En_Passant,from,position->enPassant);
				moves->eval[moves->size] = 0;
				moves->size ++;
			}
		}
	}
	pieces = position->piece[Knight+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		attack = attackKnight[from];
		dest = attack & position->piece[opponent];
		while(dest) {
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Capture,from,to);
			moves->eval[moves->size] = pieceValue[position->type[to]] - Knight_Value;
			moves->size ++;
		}
	}
	pieces = position->piece[Bishop+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		attack = positionAttackBishop(position,from);
		dest = attack & position->piece[opponent];
		while(dest){
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Capture,from,to);
			moves->eval[moves->size] = pieceValue[position->type[to]] - Bishop_Value;
			moves->size ++;
		}
	}
	pieces = position->piece[Rook+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		attack = positionAttackRook(position,from);
		dest = attack & position->piece[opponent];
		while(dest) {
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Capture,from,to);
			moves->eval[moves->size] = pieceValue[position->type[to]] - Rook_Value;
			moves->size ++;
		}
	}
	pieces = position->piece[Queen+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		attack = positionAttackQueen(position,from);
		dest = attack & position->piece[opponent];
		while(dest) {
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Capture,from,to);
			moves->eval[moves->size] = pieceValue[position->type[to]] - Queen_Value;
			moves->size ++;
		}
	}
	from = getFirstBit64(position->piece[King+player]);
	dest = attackKing[from] & position->piece[opponent];
	while(dest) {
		to = getFirstBitAndClear64(&dest);
		moves->move[moves->size] = moveGenerate(Capture,from,to);
		moves->eval[moves->size] = pieceValue[position->type[to]] - King_Value;
		moves->size ++;
	}
	moves->next = moves->toSort = 0;
}

void staticMovesGenerate(const Position* position, Moves* moves){
	guint64 pieces, dest, free;
	guint from, to, player, opponent;

	player = position->turn;
	opponent = player^1;
	free = ~(position->piece[White]|position->piece[Black]);
	moves->size = 0;
	
	if(player == White){
		pieces = position->piece[White_Pawn];
		dest = shiftUp(pieces) & free & ~RANK_8;
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to - 8;
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
		dest = shiftUp(shiftUp(pieces) & free) & RANK_4 & free;
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to - 16;
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
		if(position->castle & WHITE_CASTLE){
			if(position->castle & WHITE_KING_CASTLE && ((F1G1&free) == F1G1)){
				moves->move[moves->size] = moveGenerate(Castle,e1,g1);
				moves->eval[moves->size] = history[e1][g1];
				moves->size ++;
			}
			if(position->castle & WHITE_QUEEN_CASTLE && ((B1C1D1&free) == B1C1D1)){
				moves->move[moves->size] = moveGenerate(Castle,e1,c1);
				moves->eval[moves->size] = history[e1][c1];
				moves->size ++;
			}
		}
	}else{
		pieces = position->piece[Black_Pawn];
		dest = shiftDown(pieces) & free & ~RANK_1;
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to + 8;
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
		dest = shiftDown(shiftDown(pieces) & free) & RANK_5 & free;
		while(dest){
			to = getFirstBitAndClear64(&dest);
			from = to + 16;
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
		if(position->castle & BLACK_CASTLE){
			if(position->castle & BLACK_KING_CASTLE && ((F8G8&free) == F8G8)){
				moves->move[moves->size] = moveGenerate(Castle,e8,g8);
				moves->eval[moves->size] = history[e8][g8];
				moves->size ++;
			}
			if(position->castle & BLACK_QUEEN_CASTLE && ((B8C8D8&free) == B8C8D8)){
				moves->move[moves->size] = moveGenerate(Castle,e8,c8);
				moves->eval[moves->size] = history[e8][c8];
				moves->size ++;
			}
		}

	}
	
	pieces = position->piece[Knight+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		dest = attackKnight[from] & free;
		while(dest) {
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
	}
	
	pieces = position->piece[Bishop+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		dest = positionAttackBishop(position,from) & free;
		while(dest) {
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
	}
	
	pieces = position->piece[Rook+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		dest = positionAttackRook(position,from) & free;
		while(dest) {
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
	}
	
	pieces = position->piece[Queen+player];
	while (pieces) {
		from = getFirstBitAndClear64(&pieces);
		dest = positionAttackQueen(position,from) & free;
		while(dest) {
			to = getFirstBitAndClear64(&dest);
			moves->move[moves->size] = moveGenerate(Normal,from,to);
			moves->eval[moves->size] = history[from][to];
			moves->size ++;
		}
	}

	from = getFirstBit64(position->piece[King+player]);		
	dest = attackKing[from] & free;
	while(dest) {
		to = getFirstBitAndClear64(&dest);
		moves->move[moves->size] = moveGenerate(Normal,from,to);
		moves->eval[moves->size] = history[from][to];
		moves->size ++;
	}
	moves->next = moves->toSort = 0;
}

void rootMovesGenerate(Position* position, RootMoves* moves){
	guint index=0;
	Move move;
	Moves gMoves;
	RootMove* rMove;
	
	gMoves.state = START;
	move = nextMoveDo(position,&gMoves);
	while(move != Empty){
		rMove = &moves->move[index];
		rMove->move = move;
		rMove->eval = 0;
		rMove->depth = 0;
		rMove->seldepth = 0;
		rMove->nps = 0;
		rMove->time = 0;
		rMove->nodes = 0;
		rMove->hash = 0;
		rMove->pv[0] = Empty;
		index++;
		moveUndo(position);
		move = nextMoveDo(position,&gMoves);
	}

	moves->size = index;
}

