/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <glib/gprintf.h>

#include "position.h"
#include "init.h"

Position* positionNew();
void positionFree(Position* position);
void positionStartPosition(Position* position);
guint positionRepeat(const Position* position);
gboolean positionIsInCheck(const Position* position);
guint positionRepeat(const Position* position);
inline void emptyPosition(Position* position);
inline void addPiece(Position* position, const guint piece, const guint square);

Position* positionNew(){
    Position * position;
    position = g_malloc(sizeof(Position));
    positionStartPosition(position);
    return position;
}

void positionFree(Position* position){
	g_free(position);
}

void positionStartPosition(Position* position){
	gint i;
	emptyPosition(position);
	for(i = a2; i <=h2; i++) addPiece(position, White_Pawn, i);
	addPiece(position, White_Rook, a1);
	addPiece(position, White_Rook, h1);
	addPiece(position, White_Knight, b1);
	addPiece(position, White_Knight, g1);
	addPiece(position, White_Bishop, c1);
	addPiece(position, White_Bishop, f1);
	addPiece(position, White_Queen, d1);
	addPiece(position, White_King, e1);
	for(i = a7; i <=h7; i++) addPiece(position, Black_Pawn, i);
	addPiece(position, Black_Rook, a8);
	addPiece(position, Black_Rook, h8);
	addPiece(position, Black_Knight, b8);
	addPiece(position, Black_Knight, g8);
	addPiece(position, Black_Bishop, c8);
	addPiece(position, Black_Bishop, f8);
	addPiece(position, Black_Queen, d8);
	addPiece(position, Black_King, e8);
	position->turn = White;
	position->castle = WHITE_CASTLE | BLACK_CASTLE;
	for(i=0; i<4; i++) position->key ^= castleKey[i];
}

inline void emptyPosition(Position* position){
	gint i,j;
	for(i=0; i<14; i++) position->piece[i] = Empty;
    for(i=0; i<64; i++) position->type[i] = Empty;
	for(i=0; i<2; i++)  position->material[i] = position->pvt[i] = 0;
    position->key = position->rotateL90 = position->rotateL45 = position->rotateR45 = 0;
	position->ply = position->turn = position->enPassant = position->castle = position->staticMoves = 0;
	position->capture = Empty;
	position->inCheck = FALSE;
    for(i=0; i<Max_Ply; i++){
        position->history[i].key = 0;
		position->history[i].move = Empty;
        position->history[i].capture = position->history[i].staticMoves = 0;
        position->history[i].castle = position->history[i].enPassant = 0;
		position->history[i].inCheck = FALSE;
		for(j=0; j<2; j++)  position->material[i] = position->pvt[i] = 0;
    }
	position->phase = 0;
	position->extend = FALSE;
}

inline void addPiece(Position* position, const guint piece, const guint square){
    guint64 bitSquare = 1ull << square;
    guint color = piece%2;
    position->material[color] += pieceValue[piece];
    position->piece[color] ^= bitSquare;
   	position->piece[piece] ^= bitSquare;
    position->rotateL90 ^= 1ull << rotateL90[square];
    position->rotateL45 ^= 1ull << rotateL45[square];
    position->rotateR45 ^= 1ull << rotateR45[square];
	position->key ^= squareKey[piece][square];
    position->type[square] = piece;
    position->pvt[Opening] += pvt[piece][square][Opening];
    position->pvt[Endgame] += pvt[piece][square][Endgame];
	position->phase += piecePhase[piece]; 
}

void positionPrint(const Position* position){
	gint piece, index;
	gchar string[72] = {
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\n',
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\n',
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\n',
		'#', ' ', '#', ' ', '#', ' ', '#', ' ', '\n',
		' ', '#', ' ', '#', ' ', '#', ' ', '#', '\0'
	};
	for(index = 0; index<64; index++){
		piece = position->type[index];
		if(piece >= Pawn) string[70-rank(index)*9-(7-file(index))] = pieceToChar[piece];
	}
	g_printf("%s\n\n",string);
	if(position->turn == White) g_printf("Player: White\n");
	else g_printf("Player: Black\n");
	g_printf("Ply:%d\n",position->ply);
	g_printf("En passant:%d\n",position->enPassant);
	g_printf("Capture:%c\n",pieceToChar[position->capture]);
	g_printf("Castling:%d\n",position->castle);
	g_printf("Check:%d\n",position->inCheck);
	g_printf("Static moves:%d\n",position->staticMoves);
	g_printf("Repeat:%d\n",positionRepeat(position));
	g_printf("Material:%d\n",position->material[White]-position->material[Black]);
	g_printf("Phase:%d\n",position->phase);
	g_printf("PVT(opening/endgame):%d/%d\n",position->pvt[Opening],position->pvt[Endgame]);
	g_printf("Key:%llu\n",(long long unsigned int)position->key);
/*	g_printf("Rotate left 90:\n");
	chessBitPrint(position->rotateL90);
	g_printf("Rotate left 45:\n");
	chessBitPrint(position->rotateL45);
	g_printf("Rotate right 45:\n");
	chessBitPrint(position->rotateR45);*/
	g_printf("History: ");
	for(index=0; index<position->ply; index++){
		moveToString(position->history[index].move,string);
		g_printf("%s ",string);
	}
	g_printf("\n");
	fflush(stdout);
}

guint positionRepeat(const Position* position){
	guint i, rp = 0;
	for(i=4; i<=position->staticMoves; i+=2){
		if(position->key == position->history[position->ply-i].key){
			rp ++;
			i+=2;
		}
	}
	return rp;	
} 

void positionSetFen(Position* position, const gchar fen[]){
	guint index = 0, line = 7, square = 56, intc;
	emptyPosition(position);
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	while(fen[index] != ' '){
		intc = (gint)fen[index];
		if(intc >= 48 && intc <=59){ square += intc - 48;}
		else if(fen[index] != '/'){ addPiece(position,charToPiece(fen[index]),square++); }
		else square = --line * 8;
		index ++;
	}
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	if(fen[index] == 'b') {
		position->turn ^= 1; 
		position->key ^= turnKey;
	}
	index ++;
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	while(fen[index] != ' '){
		switch (fen[index]){
			case 'K':{ position->castle |= WHITE_KING_CASTLE; break; }
			case 'Q':{ position->castle |= WHITE_QUEEN_CASTLE; break; }
			case 'k':{ position->castle |= BLACK_KING_CASTLE; break; }
			case 'q':{ position->castle |= BLACK_QUEEN_CASTLE; break; }
		}
		index++;
	}
	index ++;
	while(fen[index] == ' ' || fen[index] == '\t') index ++;
	if(fen[index] != '-') position->enPassant = stringToSquare(fen+index);
	position->inCheck = positionIsInCheck(position);
	/* Manjkata dva podatka */
}

gboolean positionIsInCheck(const Position* position){
	guint player = position->turn;
	guint opponent = player ^ 1;
	guint square = getFirstBit64(position->piece[King+player]);
	if(positionAttackRook(position,square)&(position->piece[Rook+opponent]|position->piece[Queen+opponent]))
		return TRUE;
	if(positionAttackBishop(position,square)&(position->piece[Bishop+opponent]|position->piece[Queen+opponent]))
		return TRUE;
	if(attackKnight[square]&position->piece[Knight+opponent]) return TRUE;
	if(attackPawn[player][square]&position->piece[Pawn+opponent]) return TRUE;
	return FALSE;
}

void positionIsOk(const Position* position){
	gint i,type,j;
	gint material=0;
	for(i=White_Pawn; i<=White_Queen; i+=2){
		material += numberOfBits64(position->piece[i])*pieceValue[i];
		material -= numberOfBits64(position->piece[i+1])*pieceValue[i+1];
	}
	if(material != (position->material[White] - position->material[Black])){
		positionPrint(position);
		g_error("1. Material value is wrong (%d) !!!\n",material);
	}
	for(i=0; i<64; i++){
		type = position->type[i];
		if(type == Empty){
			for(j=White; j<=Black_King; j++){
				if((position->piece[j] & (1ull<<i)) != 0){
					positionPrint(position);
					g_error("2. Bitboard is wrong (piece=%c,square=%d) !\n",pieceToChar[j],i);
				}
				if(((1ull << rotateL90[i])& position->rotateL90) != 0){
					positionPrint(position);
					g_error("3. Rotated bitboard L90 is wrong (piece=%c,square=%d) !\n",pieceToChar[j],i);
				}
				if(((1ull << rotateL45[i])& position->rotateL45) != 0){
					positionPrint(position);
					g_error("4. Rotated bitboard L45 is wrong (piece=%c,square=%d) !\n",pieceToChar[j],i);
				}
				if(((1ull << rotateR45[i])& position->rotateR45) != 0){
					positionPrint(position);
					g_error("5. Rotated bitboard R45 is wrong (piece=%c,square=%d) !\n",pieceToChar[j],i);
				}
			}
			continue;
		}
		if((position->piece[type%2] & (1ull<<i)) == 0){
			positionPrint(position);
			g_error("6. Bitboard is wrong (piece=%c,square=%d) !\n",pieceToChar[type%2],i);
		}
		if((position->piece[(type%2)^1] & (1ull<<i)) != 0){
			positionPrint(position);
			g_error("7. Bitboard is wrong (piece=%c,square=%d) !\n",pieceToChar[type%2],i);
		}
		if((position->piece[type] & (1ull<<i)) == 0){
			positionPrint(position);
			g_error("8. Bitboard is wrong (piece=%c,square=%d) !\n",pieceToChar[type],i);
		}
		if(((1ull << rotateL90[i])& position->rotateL90) == 0){
			positionPrint(position);
			g_error("9. Rotated bitboard L90 is wrong (piece=%c,square=%d) !\n",pieceToChar[type],i);
		}
		if(((1ull << rotateL45[i])& position->rotateL45) == 0){
			positionPrint(position);
			g_error("10. Rotated bitboard L45 is wrong (piece=%c,square=%d) !\n",pieceToChar[type],i);
		}
		if(((1ull << rotateR45[i])& position->rotateR45) == 0){
			positionPrint(position);
			g_error("11. Rotated bitboard R45 is wrong (piece=%c,square=%d) !\n",pieceToChar[type],i);
		}
		for(j=White_Pawn; j<=Black_King; j++){
			if(j == type) continue;
			if((position->piece[j] & (1ull<<i)) != 0){
				positionPrint(position);
				g_error("12. Bitboard is wrong (piece=%c,square=%d) !\n",pieceToChar[j],i);
			}
		}
	}
	if(position->piece[White_King] == 0){
		positionPrint(position);
		g_error("13. Bitboard is wrong (without white king) !\n");
	}
	if(position->piece[Black_King] == 0){
		positionPrint(position);
		g_error("14. Bitboard is wrong (without black king) !\n");
	}
}

gint moveMaterial(const Position* position, const Move move){
	switch(moveGetType(move)){
		case Normal:		{ return 0; }
		case Capture:		{ return pieceValue[position->type[moveGetTo(move)]]; }
		case En_Passant:	{ return Pawn_Value; }
		case Castle:		{ return 0; }
        case Promote_Queen:	{ return Queen_Value - Pawn_Value; }
        case Promote_Rook:	{ return Rook_Value - Pawn_Value; }
        case Promote_Bishop:{ return Bishop_Value - Pawn_Value; }
        case Promote_Knight:{ return Knight_Value - Pawn_Value; }
		default: {g_error("Wrong move type (move material)!\n"); return 0;}
	}
}
