/***************************************************************************
 *   Copyright (C) 2006 by Borko Boskovic,                                 *
 *   borko.boskovic@uni-mb.si                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "startSearch.h"
#include "moveGenerator.h"
#include "search.h"

#define Game_Moves 40

void searchParameters(const Position* position, SearchParameter* sp);
void searchTimerStart(guint time);
void searchTimerStop();
guint getSearchTime(const Position* position, const SearchParameter* sp);
gpointer searchThreadStart(gpointer x);

static GThread* searchThread = NULL;
static guint stopTimer = 0;
static guint searchTime = 0;
static guint maxDepth = 0;
static guint64 maxNodes = 0;

void searchStart(Position* position, SearchParameter* searchParameter){
	stopFlag = FALSE;
	searchParameters(position,searchParameter);
	searchThread = g_thread_create(searchThreadStart,(gpointer)position,TRUE,NULL);
}

gboolean searchStop(){
	stopFlag = TRUE;
	if(!ponderSearch && searchThread != NULL){
		   g_thread_join(searchThread);
		   searchThread = NULL;
	}
	stopTimer = 0;
	return FALSE;
}

void searchParameters(const Position* position, SearchParameter* sp){
	if(sp->nodes == 0) maxNodes = G_MAXUINT64;
	else maxNodes = sp->nodes;
	if(sp->movestogo == 0) sp->movestogo = Game_Moves;
	searchTime = getSearchTime(position, sp);
	if((sp->depth == 0 && (searchTime != 0 || maxNodes != G_MAXUINT64)) || sp->infinite) 
			maxDepth = Max_Depth;
	else maxDepth = sp->depth;
	if(maxDepth == 0) maxDepth = 1;
}

gpointer searchThreadStart(gpointer position){
	if(searchTime != 0) searchTimerStart(searchTime); 
	iterativeDeepening((Position*)position,maxNodes,maxDepth);
	searchTimerStop();
	return NULL;
}

void searchTimerStart(guint time){
	if(stopTimer == 0) stopTimer = g_timeout_add(time, searchStop, NULL);
	else g_error("Timer is alredy in use!\n");
}

void searchTimerStop(){
	if(stopTimer != 0){
		if(!g_source_remove(stopTimer))
			g_error("Timer source can not remove!\n");
		stopTimer = 0;
	}
}

guint getSearchTime(const Position* position, const SearchParameter* sp){
	gint cwtime, cbtime;
	
	if(sp->movetime != 0) return sp->movetime;
	
	cwtime = (sp->wtime + sp->winc * (sp->movestogo-1)) / sp->movestogo ;
	cbtime = (sp->btime + sp->binc * (sp->movestogo-1)) / sp->movestogo ;
	cwtime = cwtime * 1.3;
	cbtime = cbtime * 1.3;

	if(position->turn == White){
		cwtime = cwtime * (cwtime/(double)cbtime);
		if(cwtime > sp->wtime || sp->movestogo <= 10) cwtime = (sp->wtime * 0.95) / sp->movestogo;
		if(cwtime == 0) cwtime = sp->wtime;
		return cwtime;
	}
	else{
		cbtime = cbtime * (cbtime/(double)cwtime);
		if(cbtime > sp->btime || sp->movestogo <= 10) cbtime = (sp->btime * 0.95) / sp->movestogo;
		if(cbtime == 0) cbtime = sp->btime;
		return cbtime;
	}
}

